/*
 *  automaton.cc created on Mon,  7 Jun 2004 20:09:45 +0200 by noe
 */    

#include "automaton.h"
#include <assert.h>

#define DETERMINISTIC(s,a) (this->_states[s].next[a].size() == 1)


#define MATCHES_AB(a,b)    (matchingmatrix[(a)][(b)])
#define SCORES_AB(a,b)     (scoringmatrix[(a)][(b)])

/*
 * output method for the current automaton
 */

ostream& operator<<(ostream& os, automaton& automaton) { 
  os << "nbstates : " << automaton._states.size()  << endl;
  /* display each state */
  for(int i=0; i<(int)automaton._states.size(); i++){
    os << "state [" << dec <<  i << "]  X:" << automaton._states[i].X  << " t:" << automaton._states[i].t << " k:" << automaton._states[i].k << endl;
    /* display forward transitions */
    for(int a=0; a < gv_align_alphabet_size ; a++){
      os << "nbtransitions on a = " << dec << a << " : " << dec << automaton._states[i].next[a].size()  << endl;
      slist<transition> & list_tran = automaton._states[i].next[a];
      for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
	os << "transition [" << dec <<  a << "] = " << dec <<  (list_iter->state) << " pr: " << (list_iter->prob)<< endl;
      }
    }
    /* display backward transitions */
  }
  return os;
}



/*
 * input method to build an automaton from a file
 */

istream& operator>>(istream& is, automaton& automaton)  {
  return is;
}


/*
 * "dot" format output (use "graphviz\dotty" or "graphviz\dot file.dot -T ps -o file.eps" to visualise it)
 */


void automaton::dot(ostream& os) { 

  os << "digraph A {"  << endl;
  os << "\t fontsize =\"10\"" << endl;
  
  /* display each state */
  for(int i = 0;i<(int)this->_states.size();i++) {
    state &s = this->_states[i];
    os << "\t \"node" << i << "\" [ label = \"" << i << ":<" << s.X << "," << s.t << "," << s.k <<">\" ];"<<endl;
  }
  
  /* display forward transitions */
  for(int i = 0;i<(int)this->_states.size();i++) { 
    state &s = this->_states[i];
    for(int a=0; a < gv_align_alphabet_size ; a++) {
      slist<transition> & list_tran = s.next[a];
      for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
	os << "\t \"node" << i <<"\"  -> \"node" << (list_iter->state) << "\" [label = \"" << a << " (" << (list_iter->prob) <<")\"] ;" << endl;
      }  
    }
  }
  os << "}"  << endl;
}



/*
 * This method create a new state with value "x"
 * and return its number.
 */

int automaton::addNewState(int X,int t,int k) {
  ++_nbStates;
  _states.push_back(state(X,t,k));
  return _nbStates;
}


/*
 * This method add a transition between two states
 * on letter "a"
 */
    
int automaton::addNewTransition(int a, int startingState, int endingState,double prob) {
    
#ifdef ASSERTB
    if (a < 0 || a >= gv_align_alphabet_size) {
      ERROR("addNewTransition","transition letter out of range");
    }
    
    if (startingState > _nbStates || endingState > _nbStates) {
      ERROR("addNewTransition","state required does not exist");
    }
#endif
 
    /* forward linking */
#ifdef ASSERTB
    // do not add twice the same state in the "a" backward list
    slist<transition>   & list_tran      = _states[startingState].next[a];
    for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
      if ( list_iter->state == endingState) {
	cerr << "> when linking state q2:" << dec << endingState << " with state q1:" << dec << startingState << " on letter a:" << dec << a << endl;
	ERROR("addNewTransition"," state q2 has already a transition on \"a\" ");
      }
    }
#endif    
    _states[startingState].next[a].push_front(transition(endingState,prob));

    /* backward linking */  
#ifdef ASSERTB
    // do not add twice the same state in the "a" backward list
    slist<transition>   & list_tran_back      = _states[endingState].prev[a];
    for(slist<transition>::iterator list_iter = list_tran_back.begin() ; list_iter != list_tran_back.end() ; list_iter++) {
      if ( list_iter->state == startingState) {
	cerr << "> when backlinking state q2:" << dec << endingState << " with state q1:" << dec << startingState << " on letter a:" << dec << a << endl;
	ERROR("addNewTransition"," state q2 has already a backtransition on \"a\" ");
      }
    }
#endif
    _states[endingState].prev[a].push_front(transition(startingState,prob));
    return 0;
}


/*
 * This method modifies the transition probability (a,startingState,endingState)
 */

    
int automaton::changeTransitionProb( int a, int startingState, int endingState, double prob) {

#ifdef ASSERTB
    if (a < 0 || a >= gv_align_alphabet_size) {
      ERROR(":changeTransitionProb","transition letter out of range");
    }
    
    if (startingState > _nbStates || endingState > _nbStates) {
      ERROR(":changeTransitionProb","state required does not exist");
    }
#endif


    /* forward link pr */
    slist<transition>   & list_tran      = _states[startingState].next[a];
    for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
      if ( list_iter->state == endingState) {
	list_iter->prob = prob;
	break;
      }
    }
   
    /* backward link pr */  
    slist<transition>   & list_tran_back      = _states[endingState].prev[a];
    for(slist<transition>::iterator list_iter = list_tran_back.begin() ; list_iter != list_tran_back.end() ; list_iter++) {
      if ( list_iter->state == startingState) {
	list_iter->prob = prob;
	break;
      }
    }
    return 0;
}



/*
 * Give the current probability of transition
 */

double automaton::Pr( int a, int startingState, int endingState) {
  
  /* forward link pr */
  slist<transition>   & list_tran      = _states[startingState].next[a];
  for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
    if ( list_iter->state == endingState) {
      return list_iter->prob;
    }
  }
  return 0.0;
}



/*
 * This method selfloops the given state on all letters
 */

void automaton::selfLoop(int stateNb) {
  for(int a=0; a<gv_align_alphabet_size; a++)
    addNewTransition(a,stateNb,stateNb,(+1e+0)/gv_align_alphabet_size);
}




/*
 * checks if there is at least on transition startingState --(a)--> ...
 */

bool automaton::hasTransition( int a, int startingState) {
#ifdef ASSERTB
  assert(startingState >= 0);
  assert(startingState < _nbStates);
  assert(a >= 0);
  assert(a < gv_align_alphabet_size);
#endif
  return _states[startingState].next[a].size() > 0;
}




#define USEQUEUEPRODUCT

/*
 * This method does the product of two automatons :
 * you can specify
 *     - "crossFinalStates"    if you need final states to be the product of BOTH automata final states : 
 *       true : automata "intersection" 
 *       false : automata this is "union".
 *     - "otherIsProbabilist"  if the second automaton is a probabilistic one
 */


automaton * automaton::product(automaton & other, bool crossFinalStates, bool otherIsProbabilist ) {
  
#ifdef ASSERTB
  if ( this->size() * sizeother > (1<<28)) {
    ERROR("product"," size of product automaton will \"certainly\" explode : better stop here ...");
  }
#endif

  automaton * result = new automaton();

#ifdef USEMAP
  typedef less< pair<int,int> > lessp;
  // for hash map__gnu_cxx::hash_map
  //typedef hash_map<pair<int,int>,int,lessp> maptype;
  // for normal map
  typedef map< pair<int,int> , int, lessp > maptype;
  maptype statesNbIndex;
#else
  vector<int> statesNbIndex( this->size() * sizeother, 0);
#endif

#ifdef USEQUEUEPRODUCT
  queue<int>  statesNbRemaining;
#else	
  stack<int>  statesNbRemaining;
#endif
  
    /* (1) start the product init state */
  pair<int,int> v                = pair<int,int>(1,1);
  int stateNumber                = result->addNewState(/*X*/1,/*t*/1);
  statesNbIndex[v]               = stateNumber;
  statesNbRemaining.push(stateNumber);
  
  /* (2) take all the non-considered interesting cases pairs on both automatons */
  while(!statesNbRemaining.empty()) {
    
    // current state remaining
#ifdef USEQUEUEPRODUCT
    int stateN =  statesNbRemaining.front();
#else	
    int stateN =  statesNbRemaining.top();
#endif
    statesNbRemaining.pop();
   
    int stateNA = result->_states[stateN].X;
    int stateNB = result->_states[stateN].t;
    pair<int,int> indexN  = pair<int,int>(stateNA, stateNB); 
    
    for(int a=0;a<gv_align_alphabet_size;a++) {
      
      slist<transition> & listA =  this->_states[stateNA].next[a];
      slist<transition> & listB =  other._states[stateNB].next[a];
      
      for(slist<transition>::iterator list_iterA = listA.begin() ; list_iterA != listA.end() ; list_iterA++) {
	for(slist<transition>::iterator list_iterB = listB.begin() ; list_iterB != listB.end() ; list_iterB++) {
	  
	  int stateAnext = list_iterA->state;
	  int stateBnext = list_iterB->state;
	  pair<int,int> indexNx  = pair<int,int>(stateAnext, stateBnext); 
	
	  int stateNx    = 0;
      
	  if ( ((stateAnext || stateBnext) && crossFinalStates) || ( stateAnext && stateBnext && (!crossFinalStates)) ) { 
	
	    if  (!statesNbIndex[indexNx]) {
	  
	      /* create a new state*/
	      stateNx = result->addNewState(stateAnext,stateBnext);
	      statesNbIndex[indexNx] = stateNx;
	      statesNbRemaining.push(stateNx);
#ifdef BUILD		
	      cout << "$push state:" << dec << stateNx << endl;
#endif	
	    } else {
	      stateNx = statesNbIndex[indexNx];
	    }
	  }  
#ifdef BUILD
	  cout << "> add transition ( a:" << dec << a << ", q1:" << dec << stateN << " , q2:" << stateNx << " ) " << endl;
#endif		
      // add a transition on from stateN --a--> stateNx.
	  if(otherIsProbabilist)
	    result->addNewTransition(a,stateN,stateNx,list_iterB->prob);
	  else
	    result->addNewTransition(a,stateN,stateNx);
	}// for (listB)
      }// for (listA)
    }// for (a)
  }//while stack nonempty

  /* Free unused data needed to build the automaton */
  statesNbIndex.clear();    
  return result;
}




/*
 * This method builds a simple Bernoulli
 * probabilistic automaton given a probability
 * matrix
 */


int automaton::Automaton_Bernoulli(vector <double> & p /*[gv_align_alphabet_size]*/) { 
  //FIXME : assert(p.size == gv_align_alphabet_size);
  int state = addNewState();
  for(int a=0; a<gv_align_alphabet_size; a++) {
    addNewTransition(a, state, state,p[a]);
  }
  return 0;
}





/** @brief build a Markov Chain of order (k)
 *  @param probabilities an array representing the probability of each A^(k+1) word
 *  @param k order of markov model
 */
int automaton::Automaton_Markov(vector<double> & p, int k ){
  
  int ApowK     = 1;
  int ApowKplus = gv_align_alphabet_size;

  /* [A] build the trie */
  int nextbase  = 1;
  addNewState();    /* starting at state 1 */


  for(int d = 1 ; d <= k ; d++, ApowK *= gv_align_alphabet_size, ApowKplus *= gv_align_alphabet_size) {

    nextbase += ApowK;

    /* [a] compute weights */
    /* (a.1) denumerators */
    vector<double> probasum  = vector<double>(ApowK,    0.0);
    int i = 0; 
    for(int buklet = 0 ; buklet < ApowK ; buklet++) {
      for(int ibuklet = 0 ; ibuklet < (int)p.size() ; ibuklet += ApowK) {
	probasum[buklet] += p[i++];
      }
    }
    /* (a.2) numerators */
    vector<double> proba     = vector<double>(ApowKplus,0.0);
    int j = 0;
    for(int buklet = 0 ; buklet < ApowKplus ; buklet++) {
      for(int ibuklet = 0 ; ibuklet < (int)p.size() ; ibuklet += ApowKplus) {
	proba[buklet]    += p[j++];
      }
    }
    
    /* [b] set states an transitions */
    for(int buklet = 0 ; buklet < ApowKplus ; buklet++) {
      int state = addNewState();
      int base  = state - 2 ;
      addNewTransition(base%gv_align_alphabet_size,
		       base/gv_align_alphabet_size + 1,
		       state,
		       proba[buklet]/probasum[buklet/gv_align_alphabet_size]);
    } /* buklet */
  } /* depth */
  
  
  /* [B] add final transitions */
  for(int b = 0 ; b < (int)(ApowK) ; b ++) {
    int bplus = b * gv_align_alphabet_size;
    double psum = 0.0;
    for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
      psum += p[bplus + a];
    }
    
    for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
      addNewTransition(a , 
		       nextbase +  (b        )%ApowK /* previous word */,
		       nextbase +  (bplus + a)%ApowK /* next word */,
		       p[bplus + a]/psum);
    }
  }
  
  return 0;
}



/*
 * This method builds a (Brejova model) M3 
 * probabilistic automaton given a probability
 * matrix
 *
 *   --> [state 1] --> [state 2] --> [state 3] -
 *  |               A             A             |
 *  |                                           |
 *   -------------------------------------------
 *                       A
 */

int automaton::Automaton_BrejovaM3(vector< vector<double> > & p /* [3][gv_align_alphabet_size] */) { 
  addNewState();
  for(int nbstates = 0 ; nbstates < 3 ; nbstates ++) {
    addNewState();
    for(int a=0; a<gv_align_alphabet_size; a++) {
      addNewTransition(a, (nbstates)%3+1,(nbstates+1)%3+1,p[nbstates][a]);
    }
  }
  return 0;
}


/*
 * this method build a (Brejova equivalent M8) M14
 * probabilistic automaton given a probability
 * of each word of size 3 (ALPHABET^3)
 */

#define M13LETTERWORDSIZE  (gv_align_alphabet_size*gv_align_alphabet_size*gv_align_alphabet_size)
#define M13SIZE            ((M13LETTERWORDSIZE-1)/2)

int automaton::Automaton_BrejovaM13(vector<double> & p /*[M13LETTERWORDSIZE]*/) {

#ifdef ASSERTB
  double pr = 0;
  for (int i=0; i<M13LETTERWORDSIZE ; i++){
    pr += p[i];
  }
  cout << "M13 : pr sum = " << pr << endl;
#endif	

  int firstState   = addNewState();
  int currentState = firstState;
  int wpivot       = M13LETTERWORDSIZE;
 
  for(int i = 0 ; i < 3 ; i++) { 
    
    double num = 0, denom = 0; 
    int    a   = 0;
    wpivot    /= gv_align_alphabet_size;

    for(int w=0; w<M13LETTERWORDSIZE; w++) { 
      
      /* compute denominator */
      if (!(w%(wpivot*gv_align_alphabet_size))) {
	a     = 0;
	denom = 0;
	for (int wp=w; wp < w + wpivot*gv_align_alphabet_size; wp++)
	  denom += p[wp];
      }
      
      /* compute numerator */
      num   += p[w];
     
      /* update result */
      if (!((w+1)%wpivot)) {	
	
	(i == 2) ?  currentState++ : currentState = addNewState();
	
	addNewTransition(a,
			 (currentState+2-firstState)/gv_align_alphabet_size+firstState-1,
			 (i == 2) ? firstState : currentState,
			 num/denom
			 );
	num = 0;
	a++;
      } 
    }
  }
  return 0;
}

/*
 * this method buid a Brejova equivalent HMM automaton 
 * ptable gives several ("nbq" more exactly) M13 models.
 * qtable gives HMM state transition matrix (size "nbq" x "nbq").
 * nbq    indicates the number of hidden state. 	
 *
 */


#define QTABLE(i,j) (qtable[(i)*nbq+(j)])
int automaton::Automaton_HMM(vector< vector<double> > & ptable /*[M13LETTERWORDSIZE][nbq]*/ , vector< vector<double> > & qtable /* [nbq][nbq] */, int nbq) {

#ifdef ASSERTB	
  // check qtable consistency
  for(int i=0;i<nbq;i++){
    double pr = 0;
    for(int j=0;j<nbq;j++){
      pr += qtable[i][j];
    }
    cout << "H(" << i <<") : pr sum = "<< pr << endl;
  }
#endif
  
  // Create the M13 automatons
  for(int i=0;i<nbq;i++){
    Automaton_BrejovaM13(ptable[i]);
  }

  // Add non deterministic transitions on the first M13 states

  // memorise old probabilities
  double * p = new double[gv_align_alphabet_size * nbq];
  for(int a=0; a<gv_align_alphabet_size; a++){
     for(int i=0; i<nbq; i++){  
       p[a  + i* gv_align_alphabet_size] = Pr(a, 1+M13SIZE*i,1+M13SIZE*i+(a+1));
     }
  }
  
  for(int a=0; a<gv_align_alphabet_size; a++){
    for(int i=0; i<nbq; i++){  
      for(int j=0; j<nbq; j++){
	if (i == j)
	  changeTransitionProb( a, 1+M13SIZE*i,  1+M13SIZE*j+(a+1), qtable[i][j] * p[a + j* gv_align_alphabet_size]);
	else
	  addNewTransition( a, 1+M13SIZE*i,  1+M13SIZE*j+(a+1), qtable[i][j] *  p[a + j* gv_align_alphabet_size]);
      }
    }
  }
  delete[] p;
  


#ifdef ASSERTB	
  for(int i=1;i<=M13SIZE*nbq;i++){
    double pr = 0;
    for (int a=0; a<gv_align_alphabet_size ; a++) {
      slist<transition> & list_tr = _states[i].next[a];
      for( slist<transition>::iterator list_iter = list_tr.begin() ;  list_iter != list_tr.end() ; list_iter++){
	pr += list_iter->prob;
      }
    }
    cout << pr << endl;
  }
#endif	


  return 0;  
}






/*
 * this method builds an Homogeneous alignment model automaton given 
 * 
 * state 0 : accepting state
 * state 1 : stating state
 * state 2 : rejecting state
 *   /\/\
 *  /\/\/
 */
#define INDEXLH(l,s) ((l)+(s)*(maxlength))
#define LVALUE(code) ((code)%(maxlength))
#define SVALUE(code) ((code)/(maxlength))
 
int automaton::Automaton_Homogeneous(vector<int> & score /*[gv_align_alphabet_size]*/, int minscore,int maxlength) {
  vector<int> statesNbIndex(maxlength*minscore, 0);
#ifdef USEQUEUE
  queue<int>  statesNbRemaining;
#else
  stack<int>  statesNbRemaining;
#endif
  int stateInit                  = addNewState(INDEXLH(0,0));
  int stateReject                = addNewState(INDEXLH(0,0));
  selfLoop(stateReject);
  int stateAccept                = 0;
  statesNbIndex[INDEXLH(0,0)] = stateInit;
  statesNbRemaining.push(stateInit);
  
  while( !statesNbRemaining.empty() ) {
    // current state remaining	
#ifdef USEQUEUE
    int stateN =  statesNbRemaining.front();
#else
    int stateN =  statesNbRemaining.top();
#endif
    statesNbRemaining.pop();
   
    int stateS = SVALUE(_states[stateN].X);
    int stateL = LVALUE(_states[stateN].X);
    for(int a=0;a<gv_align_alphabet_size;a++) {
      int stateNx = 0;
      int stateLx = stateL + 1;
      int stateSx = stateS + score[a];
      int indexNx = INDEXLH(stateLx,stateSx);
      if ( stateLx >= maxlength || stateSx <= 0) {
	/* rejecting state */
	stateNx =  stateReject;
      } else {
  	if (stateSx >= minscore) {
	  /* accepting one */
	  stateNx =  stateAccept;
	} else {
	  stateNx = statesNbIndex[indexNx];
	  if (stateNx == 0) {
	    /* create a new state and add it into the stack */
	     stateNx                   = addNewState(indexNx);
	     statesNbIndex[indexNx]    = stateNx;               
	     statesNbRemaining.push(stateNx);
	  }  
	}
      }
      addNewTransition(a,stateN,stateNx);
#ifdef ASSERTB		
      assert(DETERMINISTIC(stateN,a));
#endif
    }
  }
  statesNbIndex.clear();
  return 0;
}


/*
 * This method generates a new minimized automaton 
 * using Hopcroft minimization
 *
 */

typedef pair<int,int> ABpair;
#define REVERSESTATE(c,a,b) (((c) == (a))?(b):(((c) == (b))?(a):(c)))

automaton * automaton::Hopcroft() {

  int nbStates = this->size();   

  vector< int >                  stclass(nbStates);  // class for a given state 
  vector< list<int> >            block(nbStates);    // list of states for a given class
  vector< list<int>::iterator >  location(nbStates); // pointer to each stat inside the given list
  vector< int >                  card(nbStates);     // cardinality of a given class
  vector< vector<int> >          SET(nbStates,vector<int>(gv_align_alphabet_size,0)); // stack membership 
  
  int nbClasses;

  
  // (1) initialization
  nbClasses = 2;
  card[0]   = 1;
  card[1]   = nbStates - 1;
  
  // first class (final state only)
  block[0]    = list<int>();
  block[0].push_front(0);
  stclass[0]  = 0;
  location[0] = block[0].begin();

  // second class (all the others)
  block[1] = list<int>();
  for ( int i=1 ; i<nbStates ; i++) {
    stclass[i] = 1;
    block[1].push_front(i);
    location[i] = block[1].begin();
  } 
  
  // initial stack 
  stack<ABpair> Set;
  for(int a = 0; a < gv_align_alphabet_size ; a++) {
    Set.push(ABpair(a,0));
    SET[0][a] = 1; 
  }
  
  /* used to count (P_a_inv)Intersert(B), states membership to P_a_inv, and B membership to intersection */
  vector<int> card_P_a_inv_Intersect_B(nbStates,0);
  vector<int> Bsplit(nbStates,0);

  // (2) main hopcroft loop  
  while(!Set.empty()) {
    //(P,a) <- First(S)
    ABpair aP = Set.top(); Set.pop();
    int a = aP.first;
    int P = aP.second;
    
    SET[P][a] = 0;

    /* (2.a) IsRefined precomputation step */   
    list<int> & Plist               = block[P];
    list<int>   Plist_a_inv         = list<int>();
    // for each state in (P)
    for(list<int>::iterator iter_Plist  = Plist.begin(); 
	                    iter_Plist != Plist.end(); 
                            iter_Plist ++ ) {

      int               Pstate            = *iter_Plist;
      slist<transition>  Pstate_list_a_inv = _states[Pstate].prev[a];

      // for each predecessor *--a-->(P)
      for(slist<transition>::iterator iter_Pstate_list_a_inv  = Pstate_list_a_inv.begin(); 
                                     iter_Pstate_list_a_inv != Pstate_list_a_inv.end(); 
                                     iter_Pstate_list_a_inv ++ ) {
	
	int Pstate_a_inv = iter_Pstate_list_a_inv->state;
	int b = stclass[Pstate_a_inv];
	Plist_a_inv.push_back(Pstate_a_inv);
	card_P_a_inv_Intersect_B[b]++;
      }// transitions
    }// iter_Plist



    int oldNbClasses  = nbClasses;
    vector<int> Bprev = vector<int>();

    /* (2.b) foreach a-1(P) state */
    for(list<int>::iterator i  = Plist_a_inv.begin();
	                    i != Plist_a_inv.end();
         	            i++){
      int Pstate_a_inv = *i;
      int b            = stclass[Pstate_a_inv];
      /* if b is refined by a-1(P)*/

      if ( card_P_a_inv_Intersect_B[b] > 0 && card[b] - card_P_a_inv_Intersect_B[b] > 0) {
	
	
	int bp;
	if ( Bsplit[b] != 0 ) {
	  /* a previous state has been already moved from b to b'=Bsplit[b] */
	  bp = Bsplit[b];
	} else {
	  /* first state to be splited from b */
	  block[nbClasses]  = list<int>();
	  bp = Bsplit[b] = nbClasses;
	  Bprev.push_back(b);
	  nbClasses++;
	}
	
	block[b].erase(location[Pstate_a_inv]);
	block[bp].push_front(Pstate_a_inv);
	location[Pstate_a_inv] = block[bp].begin();
	card[bp]++;
	stclass[Pstate_a_inv] = bp;
	
      } else {
	
	card_P_a_inv_Intersect_B[b] = 0;
	
      }
    } // for each a-1(P)
    Plist_a_inv.clear();

    /* (2.c) foreach Bp state */
    // Update(S,B,B',B'')
    int i = 0;
    for(int bp = oldNbClasses ; bp < nbClasses; bp ++) {
      int b = Bprev[i++];
      /* reset tables */
      Bsplit[b] = 0;
      card[b]  -= card[bp];
      card_P_a_inv_Intersect_B[b] = 0;
      
      for(int a=0; a < gv_align_alphabet_size ; a++) {
	
	if (SET[b][a]) {
	  SET[bp][a] = 1;
	  Set.push(ABpair(a,bp));
	} else {
	  if (card[b] > card[bp]) {
	    SET[bp][a] = 1;
	    Set.push(ABpair(a,bp));
	  } else { 
	    SET[b][a] = 1;
	    Set.push(ABpair(a,b));
	  }
	}
      }
    }
    Bprev.clear();
    
  }//while (!empty)


  /* Clear Some Hopcroft data */
  Bsplit.clear();
  card_P_a_inv_Intersect_B.clear();
  location.clear();
  card.clear();
  SET.clear();

  automaton * result = new automaton();
  for(int c=1; c < nbClasses; c++) {
    result->addNewState(c);
    /* erase au states but the first */
  }
  
  
  for(int c=1; c < nbClasses; c++) {
    for(int a=0 ; a<gv_align_alphabet_size; a++) {
#ifdef ASSERTB	
      assert(DETERMINISTIC(*(block[c].begin()),a));
#endif	
      result->addNewTransition(a,  REVERSESTATE(c,stclass[1],1) , REVERSESTATE(stclass[(_states[*(block[c].begin())].next[a].begin()->state)],stclass[1],1));      
    }
  }
  
  stclass.clear();
  block.clear();
  return result;
}




/*
 * This method builds a Seed automaton on a seed
 * given at a string in the '#','@','_' alphabet
 *
 */

int automaton::Automaton_SeedPrefixesMatching(seed& s, vector< vector <int> > & matchingmatrix ) {
  
  /* Get the seed span */
  int   motif_span = s.span();
  int * motif      = s.table();
  
  
#ifdef ASSERTB  
  if (motif_span <= 0) {
    ERROR("SeedPrefixMatching","null or negative span");
  }
  
  for (int i = 0 ; i < motif_span ; i ++)
    if ( motif[i] < 0 || motif[i] >= gv_seed_alphabet_size)
      ERROR("SeedPrefixMatching","incorrect seed element");
#endif
  
#ifdef BUILD
  cout << " motif_span :" << motif_span << " , motif:" ;
  for (int i = 0 ; i < motif_span ; i++ )
    cout << motif[i] << " " ;
  cout << endl;
#endif


  /* Compute xset_bitsize and X */
  vector<int> L(motif_span+1,0);      
  int xset_bitsize  = 0; // bitsize of the set 
  int xset_size     = 0; // size of the set  
  L[0] = -1;
  for(int i=0; i<motif_span; i++) {
    if (motif[i] < (gv_seed_alphabet_size -  (gv_matching_symbol?1:0))) {
      xset_bitsize++;
      L[xset_bitsize]=i;
#ifdef BUILD
      cout << " L[" << xset_bitsize << "] : " <<  L[xset_bitsize] << endl;
#endif
    }
  }
  xset_size = 1<<xset_bitsize;
#ifdef BUILD
      cout << " xset_size :"     << xset_size << ",\t" <<
	      " xset_bitsize : " <<  xset_bitsize << endl;
#endif  
  
  /* Compute TCODE[t] to get an index by use of IndexNb = TCODE[t] + X coding system */ 
  vector<int> TCODE(motif_span+1,0);
  TCODE[0] = 0;
  for(int maxxsetsize = xset_size , l=xset_bitsize , t = 1 ; t <= motif_span; t++) {
    if (L[l] == motif_span-t) {
      maxxsetsize>>=1;
      l--;
    }
    TCODE[t] = TCODE[t-1] + maxxsetsize;
#ifdef BUILD
    cout << " TCODE["<<t<<"] : " <<  TCODE[t] << endl;
#endif	
  }
  
  
  /* 
   * 1) Precompute E,F functions
   */
  vector< vector<int> > FX(motif_span,vector<int>(gv_align_alphabet_size,0));
  vector< vector<int> > Fk(motif_span,vector<int>(gv_align_alphabet_size,0));
  vector< vector< vector<int> > > EX(motif_span, vector< vector<int> >((xset_bitsize+1), vector<int>(gv_align_alphabet_size,0)));
  vector< vector< vector<int> > > Ek(motif_span, vector< vector<int> >((xset_bitsize+1), vector<int>(gv_align_alphabet_size,0)));
  
  // F[t][a]
  for (int t=0; t<motif_span; t++) {
    for (int i = 1 ; i<= xset_bitsize && L[i] <= t ; i++) {
      int r = 1<<(i-1);
      int b = motif[L[i]];
      
      for (int a = 0 ; a<gv_align_alphabet_size ; a++){
	if (MATCHES_AB(a,b)){
	  FX[t][a] = (t>0) ? (FX[t-1][a]) | r : r; 
	  Fk[t][a] = i;
	}
      }
    }
  }
  
  
  // E[t][a]
  for (int i=1; i<=xset_bitsize; i++) { 
    int L_i = L[i]; 
    int b   = motif[L_i];
  
    for (int a=0; a<gv_align_alphabet_size; a++) {
      // does b match the mismatch a 
      if (MATCHES_AB(a,b)) {
	// if X[ib]-t-1 is a joker position, do a mask to have its bit position.
	for (int k=1; k < i; k++) {
	  int t = L[i] - L[k] - 1;
#ifdef BUILD
	  assert(t>=0);
	  assert(t<motif_span);
	  assert(k>0);
	  assert(k<=xset_bitsize);
#endif
	  EX[t][k][a] = 1<<(i-1);
	  Ek[t][k][a] = i;
	  //
	}
      }
    }
  }
  
    
  /* 2) automaton build */
  
  // fast state index (designed to retrieve a state given <X,t> code
  vector<int> statesNbIndex(TCODE[motif_span]+1 , 0);
  // queue/stack used to store non preprocessed states <X,t> code
#ifdef USEQUEUE
  queue<int>  statesNbRemaining;
#else
  stack<int>  statesNbRemaining;
#endif
  
  /* a first state [0]  has been already created : it will be the sink */
  /* create a second state [1]  : it will be the initial one */
  int Initstate_I   = addNewState(0,0,0);
  statesNbIndex[0]  = Initstate_I;
  statesNbRemaining.push( Initstate_I );
  
  while( !statesNbRemaining.empty() ) {
     
    // current state remaining
#ifdef USEQUEUE
     int Xstate_I = statesNbRemaining.front();
#else
     int Xstate_I = statesNbRemaining.top();
#endif
     statesNbRemaining.pop();
 
#ifdef ASSERTB	
     assert(Xstate_I >= 0);
     assert(Xstate_I <  (int)_states.size());
#endif	
    
     int Xstate_X = _states[Xstate_I].X;
     int Xstate_T = _states[Xstate_I].t;
     int Xstate_K = _states[Xstate_I].k;
     


#ifdef BUILD
    cout << "$pop  I:" << Xstate_I << " < X:" << Xstate_X << ", t:" << Xstate_T <<", k:"<< Xstate_K <<" > " << endl;
#endif	

#ifdef ASSERTB	
    assert(Xstate_X >= 0);
    assert(Xstate_X < xset_size);
    assert(Xstate_T >= 0);
    assert(Xstate_T < motif_span);
    assert(Xstate_K >= 0);
    assert(Xstate_K <= xset_bitsize);
    assert(TCODE[Xstate_T] + Xstate_X < TCODE[motif_span]);
#endif
    

    
    int XPstate_Ix=0, XPstate_I=0, XPstate_X=0;
    
    if (Xstate_K > 0) {
      XPstate_X   = Xstate_X ^ (1 << (Xstate_K-1));    // get X'
      XPstate_Ix  = TCODE[Xstate_T] + XPstate_X;   // get (X',t) index
      XPstate_I   = statesNbIndex[XPstate_Ix];
    }
    
    for(int a=0;a<gv_align_alphabet_size;a++) {
      
	// next state according to letter "a"
	int Ystate_X = Xstate_X;
	int Ystate_T = Xstate_T;
	int Ystate_K = Xstate_K;
      
	if ( a == (gv_align_alphabet_size - (gv_matching_symbol?1:0)) ) {
	  Ystate_T++;
        } else {

	  if ( Xstate_K > 0 ) {
#ifdef ASSERTB	
	    assert(DETERMINISTIC(XPstate_I,a));
#endif
	    int YPstate_I = _states[XPstate_I].next[a].begin()->state;
	    // YP result
	    int YPstate_X = _states[YPstate_I].X;
	    int YPstate_K = _states[YPstate_I].k;
	    //  Y result
#ifdef BUILD
	    assert(Xstate_T >= 0);
	    assert(Xstate_T < motif_span);
	    assert(Xstate_K > 0);
	    assert(Xstate_K <= xset_bitsize);
#endif

	    Ystate_X  =      YPstate_X | EX[Xstate_T][Xstate_K][a] ; // 
	    Ystate_K  = MAX( YPstate_K , Ek[Xstate_T][Xstate_K][a] );
#ifdef BUILD 
	    cout << "**V "<< endl;
	    cout << "YPstate_X:" <<  YPstate_X << ",  YPstate_K:" << YPstate_K << endl;
	    cout << "Ystate_X:"  <<  Ystate_X << ",  Ystate_K:" << Ystate_K << endl;
	  
#endif
	  } else {
	    Ystate_X  = FX[Xstate_T][a];
	    Ystate_K  = Fk[Xstate_T][a];

#ifdef BUILD 
	    cout << "*U "<< endl;
	    cout << "Ystate_X:"  <<  Ystate_X << ",  Ystate_K:" << Ystate_K<< endl;
	  
#endif

	  } // if ( Xstate_K > 0 )
	  Ystate_T  = 0;
	  
	}// if (a == gv_alphabet_size - 1)
      
	// Y result Index
	int Ystate_Ix = TCODE[Ystate_T] + Ystate_X;
	int Ystate_I  = statesNbIndex[Ystate_Ix];
      
	// if this is a final state, dont create it but link its parent to the sink state [0]
	if ( Ystate_T + L[Ystate_K] >= motif_span - 1) {
	  Ystate_I = 0;
	} else {
	  // else check if the state has not been already created
	  if (!Ystate_I) {
	    // create it and push it inside the list of states which need update.
	    Ystate_I = addNewState(Ystate_X,Ystate_T,Ystate_K);
	    statesNbIndex[Ystate_Ix] =  Ystate_I;
	    statesNbRemaining.push(Ystate_I);
#ifdef BUILD
	    cout << "$push I:"<< Ystate_I << " < X:" << hex << Ystate_X << " , t:" << dec << Ystate_T << " , k:" << Ystate_K << " > " << endl;
#endif
	  }
	}
#ifdef BUILD
	cout << "> add transition ( a:" << dec << a << ", q1:" << dec << Xstate_I << " , q2:" << Ystate_I << " ) " << endl;
#endif			
	addNewTransition(a,Xstate_I,Ystate_I);
#ifdef ASSERTB	
	assert(DETERMINISTIC(Xstate_I,a));
#endif
    }// for all letter in ALPHABET.
  }// while states remain. 
  
  /* Free unused data needed to build the automaton */
  L.clear();
  FX.clear();
  Fk.clear();
  EX.clear();
  Ek.clear();
  statesNbIndex.clear();
  return 0;
}






/*
 * This method builds a Seed automaton on a seed
 * given at a string in the '#','@','_' alphabet
 *
 */

int automaton::Automaton_SeedPrefixesMatching_old(seed& s, vector< vector <int> > & matchingmatrix) {
  
  /* Get the seed span */
  int   motif_span = s.span();
  int * motif      = s.table();


#ifdef ASSERTB
  if (motif_span <= 0) {
    ERROR("SeedPrefixMatching","negative span");
  }
#endif
  
  /* Compute xset_bitsize and X */
  vector<int> X(motif_span+1,0);
  int xset_bitsize  = 0; // bitsize of the set 
  int xset_size     = 0; // size of the set  
  
  for(int t=0; t<motif_span; t++) {
    if(motif[t] < gv_seed_alphabet_size - (gv_matching_symbol?1:0)) {  
      X[xset_bitsize]=t;
      xset_bitsize++;
    }
  }
  xset_size = 1<<xset_bitsize;
  
  
  /* Compute TCODE[t] to get an index by use of IndexNb = TCODE[t] + X coding system */ 
  vector<int> TCODE(motif_span+1,0);
  for(int maxxsetsize = xset_size , l=(xset_bitsize-1) , t = 1 ; t <= motif_span; t++) {
    if (X[l] == motif_span-t) {
      maxxsetsize>>=1;
      if(l>0)
	l--;
    }
    TCODE[t] = TCODE[t-1] + maxxsetsize;
#ifdef BUILD
    cout << " TCODE["<<t<<"] : " <<  TCODE[t] << endl;
#endif	
  }
  
  
  /* 
   * 1) Precompute u,v functions
   */
  
  vector< vector<int> >          U(motif_span,vector<int>(gv_align_alphabet_size,0));
  vector< vector<vector<int> > > V(motif_span,vector< vector<int> >(xset_bitsize,
							       vector<int>(gv_align_alphabet_size,0)
							       ));
  // U[t][a]	
  for(int t=0; t<motif_span; t++) {
    for(int i = 0 ; i<xset_bitsize && X[i]<=t ; i++) {
      int  r = 1<<i;
      int  b = motif[X[i]];

      for( int a = 0 ; a < gv_align_alphabet_size ; a++){
	if (MATCHES_AB(a,b)){
	  U[t][a] = (t>0) ? (U[t-1][a]) | r : r; 
	}
      }
    }
  }
  
  
  for(int ib=0;ib<xset_bitsize;ib++) { 
    int  xb = X[ib]; 
    for(int a=0; a<gv_align_alphabet_size;a++) {
      // does X[ib] match the mismatch a        
      int b = motif[xb];
      
      if (MATCHES_AB(a,b)) {
	// if X[ib]-t-1 is a joker position, do a mask to have its bit position.
	for(int bprev=1,ibprev=0 ; ibprev < ib ; bprev<<=1,ibprev++) {
	  int t = X[ib] - X[ibprev] - 1;
	  V[t][ib][a]=bprev;
	}
      }
    }
  }
   
  int SetMask =  (1<<xset_bitsize) - 1;
  
  /* 2) automaton build */
  
  // fast state index (designed to retrieve a state given <X,t> code
  vector<int> statesNbIndex(TCODE[motif_span]+1 , 0);
  // queue/stack used to store non preprocessed states <X,t> code
#ifdef USEQUEUE
  queue<int>  statesNbRemaining;
#else
  stack<int>  statesNbRemaining;
#endif
  
  /* a first state [0]  has been already created : it will be the sink */
  /* create a second state [1]  : it will be the initial one */
  int stateInit = addNewState(0);
  statesNbIndex[/* X=0 and t=0 */0] = stateInit;
  statesNbRemaining.push(stateInit);
  
  while( !statesNbRemaining.empty() ) {
    
    // current state remaining
#ifdef USEQUEUE
    int stateN =  statesNbRemaining.front();
#else
    int stateN =  statesNbRemaining.top();
#endif
    statesNbRemaining.pop();
    int stateX = (_states[stateN].X) &  SetMask;
    int stateT = (_states[stateN].X) >> xset_bitsize;     
    
#ifdef BUILD
    cout << "$pop current stateX < X:" << stateX << " , t:" << stateT << " > " << endl;
#endif	
    
    for(int a=0;a<gv_align_alphabet_size;a++) {
      // next state according to letter "a"
      int stateXx= stateX;
      int stateTx= 0;
      
      if ( a == gv_align_alphabet_size - (gv_matching_symbol?1:0) ) {

	stateTx = stateT + 1;

      } else {
	
      	stateXx = 0;
	for(int b=1,ib=0; ib<xset_bitsize; b<<=1,ib++) {
	  if((V[stateT][ib][a]) & stateX)
	    stateXx |= b;
	}
	stateXx |= U[stateT][a];
	
      }
      
      
      //  cout << " - reading " << dec << a << " gives < X:" << hex << stateXx << " , t:" << dec << stateTx << " > " << endl;
      
      // check if this is next state is a final one
      int maxXpos = 0;
      for(int b=1<<(xset_bitsize-1),i=xset_bitsize-1 ; b>0 ; b>>=1,i-- ) {
	if (b&stateXx) {
	  maxXpos = X[i] + 1;
	  break;
	}
      }
      
      // next state 
      unsigned int indexNx = TCODE[stateTx] + stateXx/*stateXx | (stateTx<<xset_bitsize)*/;
      int stateNx = statesNbIndex[indexNx];
      
      // if this is a final state, dont create it but link its parent to the sink state [0]
      if (stateTx + maxXpos >= motif_span) {
#ifdef BUILD
	cout << "> add transition to final state ( a:" << dec << a << ", q1:" << dec << stateN << " ,  q2:0 )" << endl;
#endif
	addNewTransition(a,stateN,0);
      } else {
	// else check if the state has not been already created
	if (!stateNx) {
	  // create it and push it inside the list of states which need update.
	  stateNx = addNewState(/* unique and easily read code <t,X> */stateXx | (stateTx<<xset_bitsize));
	  statesNbIndex[indexNx] = stateNx;
	  statesNbRemaining.push(stateNx);
#ifdef BUILD
	  cout << "$push < X:" << hex << stateXx << " , t:" << dec << stateTx << " > " << endl;
#endif
	}
#ifdef BUILD
	cout << "> add transition ( a:" << dec << a << ", q1:" << dec << stateN << " , q2:" << stateNx << " ) " << endl;
#endif	
	// add a transition on from stateN --a--> stateNx.
	addNewTransition(a,stateN,stateNx);
      }
    }// for all letter in ALPHABET.
  }// while states remain. 
  
  /* Free unused data needed to build the automaton */
  X.clear();
  TCODE.clear();
  statesNbIndex.clear();
  return 0;
}





/*
 * This method builds a Seed automaton on a seed
 * given at a string in the '#','@','_' alphabet
 *
 * if penalty != NULL, seed match is considered
 * provided that the additive penalty function
 * summed on the whole sequence is <= maxpenalty
 */

#define FAIL(s)  (_states[(s)].X)
#define REST(s)  (_states[(s)].t)
#define LEVEL(s) (_states[(s)].k)

int automaton::Automaton_SeedBuhler(seed& s, vector< vector <int> > & matchingmatrix, bool nomerge) {
 
  /* Get the seed span */
  int   motif_span  = s.span();
  int * motif       = s.table();
  
  int first_final   = 0;

#ifdef ASSERTB
  if (motif_span <= 0) {
    ERROR("Automaton_SeedBuhler","negative span");
  }
#endif
  
  /* (1) build the trie */

#ifdef USEQUEUE
  queue<int>  statesNbRemaining;
#else
  stack<int>  statesNbRemaining;
#endif
  
  /* a first state [0]  has already been created : it will be the sink */
  /* create a second state [1]  : it will be the initial one */
  int stateInit = addNewState(0,0,0);
  statesNbRemaining.push(stateInit);
  
  while( !statesNbRemaining.empty() ) {
    
    // current state remaining
#ifdef USEQUEUE
    int stateN =  statesNbRemaining.front();
#else
    int stateN =  statesNbRemaining.top();
#endif
    statesNbRemaining.pop();
    int i        = LEVEL(stateN);
    REST(stateN) = gv_align_alphabet_size; 
    int StateX   = 0; // final state
    
    if (nomerge && i == motif_span - 1 && !first_final)
	first_final = size();

    int b = motif[i];
    
    for( int a=0 ; a<gv_align_alphabet_size ;a++) {
      if (MATCHES_AB(a,b)){
	  if (i < motif_span - (gv_matching_symbol?1:0) || nomerge){
	    StateX = addNewState(0,0,i+1); // add new state
	    statesNbRemaining.push(StateX);
	  }
	  addNewTransition(/*letter*/a,stateN,StateX);
	  REST(stateN)--;
      }//matches
    }//for a
  }//while states remaining
  

  /* 
   * (2) add failure links 
   */
  
  if (!first_final)
    first_final = size();
    
  /* init first states */ 
  int  r  = 1;
  FAIL(r) = r;
  for(int a=0; a<REST(r); a++){
    addNewTransition(a,r,FAIL(r));
#ifdef ASSERTB		
    assert(DETERMINISTIC(r,a));
#endif
  }
  
  for(int a=REST(r) ; a < gv_align_alphabet_size ; a++) {
#ifdef ASSERTB		
    assert(DETERMINISTIC(r,a));
#endif
    int p = _states[r].next[a].begin()->state;
    FAIL(p) = r;
  }
 
  /* init last state */
  FAIL(0) = 0;
  REST(0) = 0;

 
  /* intermediate states:  BORDER function from Aho-Corasick*/
  for(int r = 2 ; r < size(); r++){
    for(int a=0; a<REST(r); a++){
#ifdef ASSERTB		
      assert(DETERMINISTIC(FAIL(r),a));
#endif
      if (r < first_final)
	addNewTransition(a,r,_states[FAIL(r)].next[a].begin()->state);
#ifdef ASSERTB		
      assert(DETERMINISTIC(r,a));
#endif
    }
    
    for(int a=REST(r) ; a < gv_align_alphabet_size ; a++){ 
#ifdef ASSERTB			
      assert(DETERMINISTIC(r,a));
#endif
      int p = _states[r].next[a].begin()->state;
      int s = FAIL(r);
#ifdef ASSERTB			
      assert(DETERMINISTIC(s,a));
#endif
      FAIL(p) = _states[s].next[a].begin()->state;
    }
  }
  
  return first_final;
}


#define FAIL(s)  (_states[(s)].X)
#define SCORE(s) (_states[(s)].t)
#define LEVEL(s) (_states[(s)].k)


int automaton::Automaton_SeedScore(seed & s, vector< vector <int> > & matchingmatrix, vector< vector<int> > & scoringmatrix, int scoringthreehold, bool nomerge) {
 
  /* Get the seed span */
  int   motif_span  = s.span();
  int * motif       = s.table();
  int first_final   = 0;

#ifdef ASSERTB
  if (motif_span <= 0) {
    ERROR("Automaton_SeedScore","negative span");
  }
#endif
  
  /* (0) compute the suffix best score (sbs) :
   *  it gives, for a suffix of length i the best suffix score
   *  reched : thus is prefix_score[i] + bbs[i] < scoringthreehold 
   * then it is not possible to reach the require scoringthreehold.
   */
  
  vector<int> sbs(motif_span,0);
  sbs[0] = 0;
  
  /* for all suffixes length */
  for(int i = 1 ; i < motif_span ; i++){
    int b  = motif[motif_span - i];
    int bs = NINFINITY;
    /* FIXME : this part can be precomputed out of the function */
    for(int a = 0 ; a < gv_align_alphabet_size ; a++ ) {
      if (MATCHES_AB(a,b)){
	bs = MAX(bs,SCORES_AB(a,b));
      }
    }
    sbs[i] = sbs[i-1] + bs;
#ifdef BUILD
    fprintf(stderr," sbs[%d]=%d\n",i,sbs[i]);
#endif
  }
  


  /* (1) build the trie */

#ifdef USEQUEUE
  queue<int>  statesNbRemaining;
#else
  stack<int>  statesNbRemaining;
#endif
  
  /* a first state [0]  has already been created : it will be the sink */
  /* create a second state [1]  : it will be the initial one */
  int stateInit = addNewState(0,0,0);
  statesNbRemaining.push(stateInit);
  
  while( !statesNbRemaining.empty() ) {
    
    // current state remaining
#ifdef USEQUEUE
    int stateN =  statesNbRemaining.front();
#else
    int stateN =  statesNbRemaining.top();
#endif
    statesNbRemaining.pop();
    int i        = LEVEL(stateN);
    int StateX   = 0; // final state
    
    if (nomerge && i == motif_span - 1 && !first_final)
	first_final = size();

    int b = motif[i];
    
    for( int a=0 ; a<gv_align_alphabet_size ;a++) {
      if (MATCHES_AB(a,b)){
	if (SCORE(stateN) + SCORES_AB(a,b) + sbs[motif_span - (i+1)] >= scoringthreehold){
	  if (i < motif_span - 1 || nomerge){
	    StateX = addNewState(0,0,i+1); // add new state
	    statesNbRemaining.push(StateX);
	    SCORE(StateX) = SCORE(stateN) + SCORES_AB(a,b);
	  }
	  addNewTransition(/*letter*/a,stateN,StateX);
	}
      }//prefix penalty
    }//for a
  }//while states remaining
  
  sbs.clear();
  

  /* 
   * (2) add failure links 
   */
  
  if (!first_final)
    first_final = size();
    
  /* init first states */ 
  int  r  = 1;
  FAIL(r) = r;
  for(int a = 0; a < gv_align_alphabet_size; a++ ) {
    if (!hasTransition(a,r)) {
      addNewTransition(a,r,FAIL(r));
    } else {
      int p = _states[r].next[a].begin()->state;
      FAIL(p) = r;
    }
    
#ifdef ASSERTB		
    assert(DETERMINISTIC(r,a));
#endif
  }
  
  /* init last state */
  FAIL(0) = 0;
  
 
  /* intermediate states:  BORDER function from Aho-Corasick*/
  for(int r = 2 ; r < size(); r++){
    for(int a = 0 ; a < gv_align_alphabet_size; a++){
      if (!hasTransition(a,r)) {
#ifdef ASSERTB		
	assert(DETERMINISTIC(FAIL(r),a));
#endif	
	if (r < first_final)
	  addNewTransition(a,r,_states[FAIL(r)].next[a].begin()->state);
#ifdef ASSERTB		
	assert(DETERMINISTIC(r,a));
#endif
     
      } else {
	
#ifdef ASSERTB			
      assert(DETERMINISTIC(r,a));
#endif
      int p = _states[r].next[a].begin()->state;
      int s = FAIL(r);
#ifdef ASSERTB			
      assert(DETERMINISTIC(s,a));
#endif
      FAIL(p) = _states[s].next[a].begin()->state;
      }
    }
  }
    
  return first_final;
}



/*
 * Gives the probability to reach the state numbered "StateNb"
 * (the default is 0 which means "final one")	
 * after nbSteps 
 * (64 is the common value used by everyone, i'm a sheep so i use 64)
 *
 *  ^   ^
 *   + +
 *  =(o)=
 *
 */


double automaton::PrState(int StateNb, int nbSteps) {

  vector<double> v0(size(),0);
  vector<double> v1(size(),0);

  /* set initial state probability to 1.0 */
  v1[1] = 1.00;

  /* compute probability at step i provided probabilities at step i-1 */
  for(int i= 0 ; i < nbSteps ; i++) {
    if (i&0x1){
      
      for(int nbstate=0 ; nbstate < size() ; nbstate++) {
	double val =  0.00;
	for(int a=0 ; a<gv_align_alphabet_size ; a++) {
	  slist<transition>   & list_tran      = _states[nbstate].prev[a];
	  for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
	    val += (list_iter->prob) * v0[(list_iter->state)];
	  }
	}
	v1[nbstate] = val;
      }

    } else {

      for(int nbstate=0 ; nbstate < size() ; nbstate++) {
	double val =  0.00;
	for(int a=0 ; a<gv_align_alphabet_size ; a++) {
	  slist<transition>   & list_tran      = _states[nbstate].prev[a];
	  for(slist<transition>::iterator list_iter = list_tran.begin() ; list_iter != list_tran.end() ; list_iter++) {
	    val += (list_iter->prob) * v1[(list_iter->state)];
	  }
	}
	v0[nbstate] = val;
      }
    }
  }

  double result =  ((nbSteps&0x1)?(v0):(v1))[StateNb];	
  v0.clear();
  v1.clear();

  return result;
}

