//C stuff
#include <stdio.h>
//STL
#include <vector>
#include <list>
#include <utility>
#include <algorithm>
#include <iterator>
//STD
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
#include <errno.h>

using namespace std;
#include "macro.h"
#include "automaton.h"
#include "seed.h"

#define PACKAGE "iedera"
#define VERSION "1.00"
#define MAXSEEDS 32

// choose between a subset seed and a vectorized subset seed
#define SEEDAUTOMATON(autom,seed)                           \
  ( gv_vectorized ?                                         \
     (autom).Automaton_SeedScore(                           \
               (seed),                                      \
               gv_subsetseed_matching_matrix,               \
               gv_vectorizedsubsetseed_scoring_matrix,      \
               gv_vectorizedsubsetseed_scoring_threehold    \
     )                                                      \
     :                                                      \
     (autom).Automaton_SeedPrefixesMatching(                \
               (seed),                                      \
                gv_subsetseed_matching_matrix               \
     )                                                      \
   )


// not initalized
vector< vector<int> >  gv_subsetseed_matching_matrix;
vector< vector<int> >  gv_vectorizedsubsetseed_scoring_matrix;
int                    gv_vectorizedsubsetseed_scoring_threehold = 0;
int                    gv_vectorized = 0;

vector< double >       gv_bsens;
int                    gv_bsens_k = 0;
vector< double >       gv_bsel;
int                    gv_bsel_k = 0;


// parameters
int      gv_alignment_length = 16;
int      gv_minspan   = 1,   gv_maxspan   = 5;

int      gv_minimize  = 1;
int      gv_nbruns    = 0;
int      gv_motif     = 0;

seed *   gv_seeds[MAXSEEDS];
int      gv_nbseeds   = 1;


int build_default_subsetseed_matching_matrix();
int build_default_vectorizedsubsetseed_scoring_matrix();
int build_default_probabilities();

void DISPLAYMATRIX( std::ostream & os, vector< vector<int> > & matrix );


void USAGE() {
  cerr << "* USAGE:" << endl;
  cerr << "  " << PACKAGE << " [options]" << endl;
  cerr << "      -h       display this Help screen" << endl ;
  cerr << "      -v       display program Version" << endl ;
  cerr << "                                                                                      " << endl;
  cerr << "   1) Seed Model:" << endl;
  cerr << "     a) Data "<< endl;
  cerr << "      -A <int>                : align alphabet size \'A\' (default " << gv_align_alphabet_size << ")" << endl;
  cerr << "      -B <int>                :  seed alphabet size \'B\' (default " <<  gv_seed_alphabet_size << ")" << endl;
  cerr << "      -M {{<bool>}} table     : subset seed Matching matrix" << endl;
  cerr << "         (default "; DISPLAYMATRIX(cerr,gv_subsetseed_matching_matrix); cerr << " )" << endl;
  cerr << "         * note    : parameter is a C-like matrix of 'A'x'B' cols x rows " << endl;
  cerr << "         * example : -M \"{{1,0,1},{0,1,1}}\"    (for A=3 and B=2) " << endl;
  cerr << "     b) SubsetSeed/VectorizedSubsetSeed "<< endl;
  cerr << "      -V                      : select the Vectorized subset seed model (default " << gv_vectorized <<")"<< endl;
  cerr << "      -S {{<int>}} table      : vectorized subset seed Scoring table " << endl;
  cerr << "      -T <int>                : vectorized subset seed scoring Threehold " << endl;
  cerr << "                                                                               " << endl;
  cerr << "   2) Alignments:" << endl;
  cerr << "      -b <double>,<double>,...: fix the Bernoulli/Markov background distribution (size |'A'|^k)" << endl;
  cerr << "      -f <double>,<double>,...: fix the Bernoulli/Markov foreground distribution (size |'A'|^k)" << endl;
  cerr << "      -l <int>                : fix the alignment length (default = " <<  gv_alignment_length << ") " << endl;
  cerr << "                                                                               " << endl;	
  cerr << "   3) Seed Enumeration:" << endl;
  cerr << "      -s <int>,<int>          : seed span (default min=" << gv_minspan << ",max=" << gv_maxspan << ")" << endl;
  cerr << "      -n <int>                : number of seeds  (default " << gv_nbseeds << ")" << endl;
  cerr << "      -r <int>                : random enumeration  (default " << gv_nbruns << " is complete enumeration)" << endl;
  cerr << "      -w <double>,<double>    : minimal/maximal weight (default min=" << gv_minweight << ",max=" << gv_maxweight << ")" << endl;
  cerr << "         * symbol # (if defined, otherwise the max of all symbols) is of weight 1.0 " << endl;
  cerr << "         * symbol _ is of weight 0.0 " << endl;
  cerr << "         * symbol e weight is given by 'log_#(e matches background distribution)'" << endl;
  cerr << "      -i <int>,<int>,...      : signature id (number of '0','1',...'B-1' elements inside a subset seed)" << endl;
  cerr << "         * letters '0' are added to fit the seed span " << endl;
  cerr << "      -d                      : disable Hopcroft minimization (default is enabled) " << endl;
  cerr << "      -m <seeds>              : check the given seed selectivity/sensitivity " << endl;
  exit(-1);
}


void DISPLAYMATRIX( std::ostream & os, vector< vector<int> > & matrix ) {
  os << "{";
  for (int b=0; b<gv_seed_alphabet_size; b++) {
    os << "{";
    for (int a=0; a<gv_align_alphabet_size; a++) {
      os << (matrix[a][b]);
      if (a < (gv_align_alphabet_size-1)) os << ",";
    }
    os << "}";
    if (b < (gv_seed_alphabet_size-1)) os << ",";
  }
  os << "}";
}

void DISPLAYTABLE( std::ostream & os, vector<double> & table ) {
  os << "{";
  for (int a=0; a<(int)table.size(); a++) {
    os << (table[a]);
    if (a < (int)(table.size()-1))
      os << ",";
  }
  os << "}";
}


void CHECKPROB(vector<double> & table) {
  double result = 0.0;
  for(int j=0 ; j<(int)table.size() ; j++)
    result += (table)[j];
  if(result < 0.99 || result > 1.01)
    ERROR("CHECKPROB"," sum of " << (table)[0] << " + ... + " << (table)[table.size()-1] << " must be 1.0 and not " << result);
}

void VERS() {
  cerr << PACKAGE << " version " << VERSION << endl;
  exit(-1);
}

void PARSEFLAG(int & i, char ** argv, int argc, int & var, int value) {
  var = value;
}



void PARSEINT(int & i,  char ** argv, int argc, int & var, int min, int max) {
  i++;
  if (i >= argc)
    ERROR("PARSEINT","\"" << argv[i-1] << "\" found without argument");
  char * next = argv[i];
  int i_tmp = strtol(argv[i],&next,10);
  if (errno != 0 || ((*next) != '\0'))
    ERROR("PARSEINT","\"" << argv[i-1] << "\" is not followed by an integer");
  if (i_tmp>max || i_tmp <min)
    ERROR("PARSEINT","\"" << argv[i-1] << "\" is followed by an integer outside the valid range");
  var = i_tmp;
}


void PARSEDINT(int & i, char ** argv, int argc,
	       int & var1, int min1, int max1,
	       int & var2, int min2, int max2) {
  i++;
  if (i >= argc)
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" found without argument");
  int i_tmp = sscanf(argv[i],"%i,%i",&var1,&var2);
  if (i_tmp != 2 )
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" is not followed by an <int>,<int>");
  if (var1 < min1 || var1 > max1 || var2 < min2 || var2 > max2)
    ERROR("PARSEDINT","\"" << argv[i-1] << "\" is followed by integers out of the range");
}



void PARSEDOUBLE(int & i, char ** argv, int argc, double & var , double min, double max) {
  i++;
  if (i >= argc)
    ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" found without argument");
  char * next = argv[i];
  double d_tmp = strtod(argv[i],&next);
  if (errno != 0 || ((*next) != '\0'))
    ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" is not followed by a double") ;
  if (d_tmp>max || d_tmp <min)
    ERROR("PARSE","\"" << argv[i-1] << "\" is followed by a double outside the valid range");
  var = d_tmp;
}


void PARSEDDOUBLE(int & i, char ** argv, int argc,
		  double & var1, double min1, double max1,
		  double & var2, double min2, double max2) {
  i++;
  if (i >= argc)
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" found without argument");
  int i_tmp = sscanf(argv[i],"%lf,%lf",&var1,&var2);
  if (i_tmp != 2 )
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is not followed by an <dbl>,<dbl>");
  if (var1 < min1 || var1 > max1 || var2 < min2 || var2 > max2)
    ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is followed by doubles out of the range");
}



void PARSESIGNATURE(int & i, char ** argv, int argc, vector<int> & table) {
  i++;
  if (i >= argc)
    ERROR("PARSESIGNATURE","\"" << argv[i-1] << "\" found without argument");

  /* read table */
  char * pch = strtok(argv[i],",;");
  table.clear();
  while (pch){
    int value = 0;
    int i_tmp = sscanf(pch,"%d",&value);
    if (i_tmp != 1)
      ERROR("PARSESIGNATURE","\"" << pch << "\" is not a correct integer");
    if (value < 0 || value >= 64)
      ERROR("PARSESIGNATURE","\"" << pch << "\" is out of the range [0..N] (N=64)");
    table.push_back(value);
    pch = strtok(NULL,",;");
  }

  /* check table size */
  if ((int)table.size() != gv_seed_alphabet_size)
    ERROR("PARSESIGNATURE", " there is not a correct number of <int> values given by " << argv[i-1] << " when compared with the current seed alphabet size B = " << gv_seed_alphabet_size);
  
}



void PARSEPROBS(int & i, char ** argv, int argc,
		vector<double> & table, int & k) {
  i++;
  if (i >= argc)
    ERROR("PARSEPROBS","\"" << argv[i-1] << "\" found without argument");

  /* read table */
  char * pch = strtok(argv[i],",;");
  table.clear();
  while (pch){
    double value = 0.0;
    int i_tmp = sscanf(pch,"%lf",&value);
    if (i_tmp != 1)
      ERROR("PARSEPROBS","\"" << pch << "\" is not a correct <dbl>");
    if (value < 0 || value > 1 )
      ERROR("PARSEPROBS","\"" << pch << "\" is  out of the range [0,1]");
    table.push_back(value);
    pch = strtok(NULL,",;");
  }

  /* check table size */
  int a = gv_align_alphabet_size;
  k = 0;
  while(1) {
    if ((int)table.size() == a)
      break;
    else if ((int)table.size() < a)
      ERROR("PARSEPROBS", " there is not a correct number of <dbl> values ("<< (table.size()) <<") given by " << argv[i-1] << " when compared with the current alphabet size A = " << gv_align_alphabet_size);
    a *= gv_align_alphabet_size;
    k++;
  }
  CHECKPROB(table);
}







typedef enum {
  dummy,
  readfirstopenbracket,
  readsecondopenbracket,
  readint,
  readintseparator,
  readsecondclosedbracket,
  readsecondbracketseparator,
  readfirstclosedbracket,
  readfirstbracketseparator
} automatareadmatrixstates;




void PARSEMATRIX(int & i, char ** argv, int argc, vector< vector<int> > & matrix, int min, int max, int gnbcolumns, int gnbrows) {
  i++;
  if (i >= argc)
    ERROR("PARSEMATRIX","\"" << argv[i-1] << "\" found without argument");
  char * pch = argv[i];
  int nbcolumns = 0, nbrows = 0;

  automatareadmatrixstates state = dummy;
  /* automata */
  while(*pch) {

    switch (*pch){
    case '{':
      switch (state){
      case dummy:
	nbrows = -1;
	state = readfirstopenbracket;
	break;
      case readfirstopenbracket:
      case readsecondclosedbracket:
      case readsecondbracketseparator:
	nbcolumns = 0;
	nbrows ++;
	if (nbrows > gnbrows)
	  ERROR("PARSEMATRIX","\" invalid number of lines : > " << gnbrows << "in \"" << argv[i] << "\" string parameter ");
	state = readsecondopenbracket;
	break;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    case '}':
      switch (state){
      case readint:
      case readintseparator:
	state = readsecondclosedbracket;
	if (nbcolumns != gnbcolumns)
	  ERROR("PARSEMATRIX","\" invalid number of columns :  > " << gnbcolumns << " != " << gnbcolumns << " in \"" << argv[i] << "\" string parameter ");
	break;
      case readsecondclosedbracket:
	state = readfirstclosedbracket;
	if (nbrows != gnbrows - 1)
	  ERROR("PARSEMATRIX","\" invalid number of lines : " << nbrows << " != " << (gnbrows - 1) << " in \"" << argv[i] << "\" string parameter ");
	break;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    case ' ':
    case '\t':
      pch++;
      break;

    case ';':
    case ',':
      switch (state){
      case readint:
	state = readintseparator;
	break;
      case readsecondclosedbracket:
	state = readsecondbracketseparator;
	break;
      case readfirstclosedbracket:
	state = readfirstbracketseparator;
      default:
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
      pch++;
      break;

    default: /* read value */
      if (state == readsecondopenbracket || state == readintseparator) {
	state = readint;
	int value = strtol (pch, &pch, 10);
	if (value>max || value<min)
	  ERROR("PARSEMATRIX","\"" << argv[i] << "\" contains at least one integer outside the valid range");
	matrix[nbcolumns][nbrows] = value;
	nbcolumns++;
	if (nbcolumns > gnbcolumns)
	  ERROR("PARSEMATRIX","\" invalid number of columns : " << nbcolumns << " in \"" << argv[i] << "\" string parameter ");
      } else {
	ERROR("PARSEMATRIX","\" invalid \'" << *pch << "\' in \"" << argv[i] << "\" string parameter ");
      }
    }
  } /* while */
  if (state == readfirstclosedbracket || state == readfirstbracketseparator)
    return;
  else
    ERROR("PARSEMATRIX","\" unfinished string parameter \"" << argv[i] << "\" ");
}



void CHECKMATCHINGMATRIX( vector< vector<int> > & matrix ) {
  /* check the '1' symbol existance */ 
  for (int b=0; b<gv_seed_alphabet_size; b++) {
    if ( matrix[gv_align_alphabet_size-1][b] == 0 ) {
      gv_matching_symbol = false;
    }
  }
  /* check the '#' symbol existance */
  for(int a=0; a <gv_align_alphabet_size-1;a++) {
    if ( matrix[a][gv_seed_alphabet_size-1] == 1 ) {
      gv_matching_symbol = false;
    }
  }
  if ( matrix[gv_align_alphabet_size-1][gv_seed_alphabet_size-1] == 0 ) {
    gv_matching_symbol = false;
  }
}




void PARSESEEDS(int & i, char ** argv, int argc) {
  i++;
  if (i >= argc)
    ERROR("PARSESEEDS","\"" << argv[i-1] << "\" found without argument");

  gv_motif   = 1;
  gv_nbseeds = 0;

  char * pch = strtok(argv[i],",;");
  while (pch) {
    string s = string(strdup(pch));
    gv_seeds[gv_nbseeds] = new seed(s);
    gv_nbseeds++;
    if ( gv_nbseeds > MAXSEEDS)
      ERROR("PARSESEEDS","\" element " << gv_nbseeds << "\" is out of array bounds");
    pch = strtok(NULL,",;");
  }
}



void SCANARG(int argc , char ** argv) {
  for(int i=1; i<argc; i++){

    /* 1) Seed model */
    if(!strcmp(argv[i],"-A")||!strcmp(argv[i],"--A")) {
      PARSEINT(i,argv,argc,gv_align_alphabet_size,2,1024);
      build_default_subsetseed_matching_matrix();
      build_default_vectorizedsubsetseed_scoring_matrix();
      build_default_probabilities();
    } else if(!strcmp(argv[i],"-B")||!strcmp(argv[i],"--B")) {
      PARSEINT(i,argv,argc,gv_seed_alphabet_size,1,1024);
      build_default_subsetseed_matching_matrix();
      build_default_vectorizedsubsetseed_scoring_matrix();
      build_default_probabilities();
      gv_signature_flag = 0;
    } else if(!strcmp(argv[i],"-M")||!strcmp(argv[i],"--Matching")) {
      PARSEMATRIX(i,argv,argc,gv_subsetseed_matching_matrix,0,1,gv_align_alphabet_size,gv_seed_alphabet_size);
      CHECKMATCHINGMATRIX(gv_subsetseed_matching_matrix);
    } else if(!strcmp(argv[i],"-S")||!strcmp(argv[i],"--Scoring")) {
      PARSEMATRIX(i,argv,argc,gv_vectorizedsubsetseed_scoring_matrix,-10000,10000,gv_align_alphabet_size,gv_seed_alphabet_size);
    } else if(!strcmp(argv[i],"-T")||!strcmp(argv[i],"--Threehold")) {
      PARSEINT(i,argv,argc,gv_vectorizedsubsetseed_scoring_threehold,-10000,+10000);
    } else if(!strcmp(argv[i],"-V")||!strcmp(argv[i],"--Vectorized")) {
      PARSEFLAG(i,argv,argc,gv_vectorized,1);
    

    /* 2) Alignment model */
    } else if(!strcmp(argv[i],"-b")||!strcmp(argv[i],"--background")) {
      PARSEPROBS(i,argv,argc,gv_bsel,gv_bsel_k);
    } else if(!strcmp(argv[i],"-f")||!strcmp(argv[i],"--foreground")) {
      PARSEPROBS(i,argv,argc,gv_bsens,gv_bsens_k);
    } else if(!strcmp(argv[i],"-l")||!strcmp(argv[i],"--length")) {
      PARSEINT(i,argv,argc,gv_alignment_length,1,1000000);

    /* 3) Seed enumeration */
    } else if(!strcmp(argv[i],"-s")||!strcmp(argv[i],"--span")) {
      PARSEDINT(i,argv,argc,gv_minspan,1,1000,gv_maxspan,1,1000);
    } else if(!strcmp(argv[i],"-w")||!strcmp(argv[i],"--weight")) {
      PARSEDDOUBLE(i,argv,argc,gv_minweight,0,1e32,gv_maxweight,0,1e32);
    } else if(!strcmp(argv[i],"-n")||!strcmp(argv[i],"--number")) {
      PARSEINT(i,argv,argc,gv_nbseeds,1,1000);
    } else if(!strcmp(argv[i],"-r")||!strcmp(argv[i],"--random")) {
      PARSEINT(i,argv,argc,gv_nbruns,1,1000000000);
    } else if(!strcmp(argv[i],"-d")||!strcmp(argv[i],"--random")) {
      PARSEFLAG(i,argv,argc,gv_minimize,0);
    } else if(!strcmp(argv[i],"-m")||!strcmp(argv[i],"--motifs")) {
      PARSESEEDS(i,argv,argc);
    } else if(!strcmp(argv[i],"-i")||!strcmp(argv[i],"--idsign")) {
      PARSESIGNATURE(i,argv,argc,gv_signature);
      gv_signature_flag = true;
      /* 4) misc */
    } else if(!strcmp(argv[i],"-h")||!strcmp(argv[i],"--help")) {
      USAGE();
    } else if(!strcmp(argv[i],"-v")||!strcmp(argv[i],"--version")) {
      VERS();
    } else ERROR("PARSE","\"" << argv[i] << "\" is not a valid argument");
  }
}



class seedproperties {

/**
 * @class seedproperties
 * @brief simple "inner class" to keep seed properties as
 * sensitivity, selectivity ... and other properties
 */

 public:

  seedproperties(double sel, double sens, double dist, std::string str){
    this->sel  = sel;
    this->sens = sens;
    this->dist = dist;
    this->str  = string(str);
  }

  double sel;
  double sens;
  double dist;
  string str;

  /**
   * @brief comparison operator (sort on selectivity first, then on sensitivity for equal selectivity)
   */
  bool operator<(seedproperties & e) {
    return (this->sel < e.sel) || ( (this->sel == e.sel) && (this->sens < e.sens)) ;
  }


  friend std::ostream& operator<<(std::ostream& os, seedproperties& e);

};

std::ostream& operator<<(std::ostream& os, seedproperties& e){
  os <<  e.str << "\t" << e.sel <<  "\t" << e.sens <<  "\t" << e.dist ;
  return os;
}









double dist(double sel,double sens) {
  return sqrt((1.0 - sel)*(1.0 - sel) + (1.0 - sens)*(1.0 - sens));
}


/**
 * @brief display a list of seeds properties
 */

int display (list<seedproperties> & l) {
  cerr << "#" << "\t" << "1th column : seed motif"
              << "\t" << "2nd column : selectivity"
              << "\t" << "3rd column : sensitivity"
              << "\t" << "4th column : distance to (1,1)" ;

  l.sort();
  for(list<seedproperties>::iterator i = l.begin(); i != l.end() ; i++) {
    cout << (*i) << endl;
  }
  return 0;

}


/**
 * @brief Compute the convex hull area for the selectivity/sensitivity of a set of seeds
 * @brief [this definition is not suitable and must NOT be used]
 */

double areaConvex(list<seedproperties> & l) {


  /* (1) push border seedproperties and sort */
  l.sort();
  l.push_front(seedproperties(0.0,1.0,1.0,string("_")));
  l.push_back( seedproperties(1.0,0.0,1.0,string("_")));

  /* (2) compute convex hull */
  list<seedproperties>::iterator i = l.begin();
  vector<seedproperties>         c = vector<seedproperties>();
  vector<double>          s = vector<double>();
  c.push_back(*(i));
  s.push_back(+0.0);
  i++;

  while(i != l.end()) {
    /* some elments can be negative */
    if ((i->sel >= 0) && (i->sens >= 0)) {
    eval:
      /* increased x-sel */
      if (i->sel > c.back().sel){
	double slope = (i->sens - c.back().sens)/(i->sel - c.back().sel);
	
	if (slope <= s.back()){
	  c.push_back(*i);
	  s.push_back(slope);
	}else{
	  if (s.size() > 1){
	    c.pop_back();
	    s.pop_back();
	    goto eval;
	  }
	}
	
      }else{
	
	/* same x-sel but increased sens */
	if ( i->sens > c.back().sens){
	  c.pop_back();
	  s.pop_back();
	  double slope = (i->sens - c.back().sens)/(i->sel - c.back().sel);
	
	  if (slope <= s.back()){
	    c.push_back(*i);
	    s.push_back(slope);
	  }else{
	    if (s.size() > 1){
	      c.pop_back();
	      s.pop_back();
	      goto eval;
	    }
	  }
	
	}
      }
    }
    i++;
  }

  /* (3) compute covered area */
  double area = 0.0;
  vector<seedproperties>::iterator j = c.begin();
  vector<double>::iterator  k = s.begin();
  for( vector<seedproperties>::iterator i = c.begin() ; i !=  c.end() ; i++ ){
    area += 0.5 * (i->sel - j->sel) * (i->sens + j->sens) ;
    j = i;
    //cout << (*i) << endl;
  }
  cerr << "convex hull area : " << area << endl;
  return area;
}




/**
 * @brief Select seeds that are on the current "pareto set"
 * @brief usefull to keep best seeds only
 */


double selectPareto(list<seedproperties> & l) {

  /* (1) push border seedproperties and sort */
  l.sort();

  /* (2) compute border hull */
  list<seedproperties>::iterator i = l.begin();
  list<seedproperties>::iterator j = l.begin();
  i++;

  while(i != l.end()) {

    /* increased sensitivity */
    if (i->sens > j->sens /* note that i->sel >= j-> sel */) {
      if (j != l.begin()) {
	list<seedproperties>::iterator k = j;
	j--;
	l.erase(k);
      } else {
	l.erase(j);
	j = i;
	i++;
      }
    } else {
      /* increased selectivity */
      if (i->sel > j->sel) {
	j = i;
	i++;
      } else {
	/* equal selectivity but decreased sensitivity */
	list<seedproperties>::iterator k = i;
	i++;
	l.erase(k);
      }
    }
  } /* while */



  /* (3) compute covered area */
  double area = 0.0;
  j = l.begin();
  for( i = l.begin() ; i !=  l.end() ; i++ ){
    area += (i->sel - j->sel) * (i->sens) ;
    j = i;
    //cerr << (*i) << endl;
  }
  return area;
}



/**
 * @brief Compute the "pareto set" area for the selectivity/sensitivity of a set of seeds
 * @brief [this definition is more suitable]
 */

double list_and_areaPareto(list<seedproperties> & l) {


  /* (1) select pareto set and compute area */
  double area = selectPareto(l);
  cerr << "pareto area : " << area << endl;
  /* (2) display seeds */
  for( list<seedproperties>::iterator i = l.begin() ; i !=  l.end() ; i++ ){
    cout << (*i) << endl;
  }
  return area;
}






/*
 *@ brief thie method is needeed when |A| or |B| size is changed
 */

int build_default_subsetseed_matching_matrix() {
  /* build matrix */
  gv_subsetseed_matching_matrix.clear();
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    vector<int> * v = new vector<int>(gv_seed_alphabet_size,0);
    gv_subsetseed_matching_matrix.push_back(*v);
  }

  /* fill it in */
  for(int b = 0 ; b < gv_seed_alphabet_size ; b++ ) {
    /* sharp */
    gv_subsetseed_matching_matrix[gv_align_alphabet_size-1][b] = 1;
    for(int a = b; a < gv_align_alphabet_size ; a++ )
      gv_subsetseed_matching_matrix[a][b] = 1;
  }

  /* set the flag to signal the '#' element */
  gv_matching_symbol = true ; 
  
  return 0;
}




/**
 *@ brief thie method is needeed when |A| or |B| size is changed
 */

int build_default_vectorizedsubsetseed_scoring_matrix() {

  /* build matrix */
  gv_vectorizedsubsetseed_scoring_matrix.clear();
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    vector<int> * v = new vector<int>(gv_seed_alphabet_size,-1);
    gv_vectorizedsubsetseed_scoring_matrix.push_back(*v);
  }

  /* fill it in */
  for (int a = 0 ; a < gv_align_alphabet_size ; a++) {
    for(int b = 0 ; b < gv_seed_alphabet_size ; b++ ) {
      if (gv_subsetseed_matching_matrix[a][b])
	gv_vectorizedsubsetseed_scoring_matrix[a][b] = 1;
    }
  }
  return 0;
}



/**
 *@ brief thie method is needeed when |A| size is changed
 */

int build_default_probabilities()  {
  gv_bsens          = vector <double>(gv_align_alphabet_size,0);
  gv_bsens_k        = 0;
  gv_bsel           = vector <double>(gv_align_alphabet_size,0);
  gv_bsel_k         = 0;

  double remaining  = 1.0;

  /* bsel : 0.5,0.25,0.125,0.125,
   * bsens : 0.125,0.125,0.25,0.5
   */

  for(int i = 0 ; i < gv_align_alphabet_size; i++) {
    if ( i < (gv_align_alphabet_size - 1) ) {
      remaining *= .5;
    }

    gv_bsel[i] = remaining;
    gv_bsens[gv_align_alphabet_size-i-1] = remaining;
  }

  return 0;
}

void computeWeight() {
  gv_bsel_weight  = vector<double>(gv_seed_alphabet_size);
  gv_bsel_minprob =  1e32;
  gv_bsel_maxprob = -1e32;

  /* compute background probs */
  for(int b = 0; b < gv_seed_alphabet_size ; b++) {
    double pr = 0.0;
    for(int a = 0; a < gv_align_alphabet_size ; a++) {
      if (gv_subsetseed_matching_matrix[a][b]) {
	for (int i = 0; i < power(gv_align_alphabet_size,gv_bsel_k) ; i++)
	  pr += gv_bsel[a+i*gv_align_alphabet_size];
      }
    }
    gv_bsel_weight[b] = pr;
    gv_bsel_minprob   = MIN(pr,gv_bsel_minprob);
    gv_bsel_maxprob   = MAX(pr,gv_bsel_maxprob);
     
  }
  
  /* take the log_#(b) */
  for(int b = 0; b < gv_seed_alphabet_size ; b++) {
    gv_bsel_weight[b] = log(gv_bsel_weight[b])/log(gv_bsel_minprob) ;
    /* FIXME : normalize ?? */
  }
  
}





int main(int argc, char * argv[]) {

  int nbruns = 0;

  list<seedproperties> l = list<seedproperties>();
  l.push_front(seedproperties(0.0,1.0,1.0,string("_")));
  l.push_back( seedproperties(1.0,0.0,1.0,string("_")));


  build_default_subsetseed_matching_matrix();
  build_default_vectorizedsubsetseed_scoring_matrix();
  build_default_probabilities();

  SCANARG(argc,argv);

  /*
   * O) DISPLAY PARAMETERS
   */
  cerr << "* alignment alphabet size |A| : " << gv_align_alphabet_size << " , " << "seed alphabet size |B| : "  << gv_seed_alphabet_size << endl;
 cerr << "* subset seed matching matrix : "; DISPLAYMATRIX(cerr,gv_subsetseed_matching_matrix); cerr << endl;
  if (gv_vectorized) {
    cerr << "* subset seed scoring matrix : "; DISPLAYMATRIX(cerr,gv_vectorizedsubsetseed_scoring_matrix); cerr << endl;
    cerr << "* subset seed scoring threehold : " << gv_vectorizedsubsetseed_scoring_threehold << endl;
  }

  /* compute the weight for each seed element */
  computeWeight();
  cerr << "* seed alphabet weights " ; DISPLAYTABLE(cerr,gv_bsel_weight); cerr <<endl;
  /* build the foreground/background models */
  automaton a_sel;
  if (gv_bsel_k == 0) {
    a_sel.Automaton_Bernoulli(gv_bsel);
    cerr << "* background Bernoulli model "; DISPLAYTABLE(cerr,gv_bsel); cerr << endl;
  } else {
    a_sel.Automaton_Markov(gv_bsel,gv_bsel_k);
    cerr << "* background Markov model "; DISPLAYTABLE(cerr,gv_bsel);
    cerr << " (order : " << (gv_bsel_k)<< ")" << endl;
  }

  automaton a_sens;
  if (gv_bsens_k == 0) {
    a_sens.Automaton_Bernoulli(gv_bsens);
    cerr << "* foreground Bernoulli model "; DISPLAYTABLE(cerr,gv_bsens); cerr << endl;
  } else {
    a_sens.Automaton_Markov(gv_bsens,gv_bsens_k);
    cerr << "* foreground Markov model "; DISPLAYTABLE(cerr,gv_bsens);
    cerr << " (order : " << (gv_bsens_k) << ")" << endl;
  }


  /*
   * I) MAIN STATISTICAL PART
   */



  if ( gv_motif ) {
    
    /*
     * A) check a command line given set of seeds ( -m 1022,20102)
     */
    
    automaton * a_spr = NULL;
    double      selp = 0.0;

    for(int i=0 ; i < gv_nbseeds ; i++){

      /* seed automaton */
      
      automaton * a_s = new automaton();
	
      cerr << " seed : " << (gv_seeds[i]) << endl;
      SEEDAUTOMATON(*a_s,*gv_seeds[i]);
      cerr << " size of " << (gv_seeds[i]) << " : " << (a_s->size()) << endl;;

      if (gv_minimize) {
	automaton * na = a_s->Hopcroft(); 
	delete a_s; 
	a_s = na ;
      }

      cerr << " reduced size : " << (a_s->size()) << endl;;

      /* (1) compute partial selectivity */
      automaton * a_pr_s_sel      = a_s->product(a_sel, false, true);
      selp                        = MAX(selp,a_pr_s_sel->PrState(0,gv_seeds[i]->span()));
      delete a_pr_s_sel;


      /* (2) product automaton for sensitivity */

      if (i > 0) {
	automaton * a_spr_temp = a_spr;
	a_spr =  a_spr->product(*a_s, false, false);
	cerr << " product size : " << (a_spr->size()) << endl;;
      	if (gv_minimize) {
	  automaton * na = a_spr->Hopcroft(); 
	  delete a_spr; 
	  a_spr = na ;
	}
      	cerr << " reduced product size : " << (a_spr->size()) << endl;;
	delete a_spr_temp;
	delete a_s;
      } else {
	a_spr = a_s;
      }
	 
    }// for each seed


    /* (3) compute sensitivity */
    automaton * a_pr_sens     = a_spr->product(a_sens, false, true);
    double sens               = a_pr_sens->PrState(0,gv_alignment_length);	
    delete      a_pr_sens;

    double sel = 1.0 - MIN(1.0,gv_nbseeds*selp);
    
    delete a_spr;

    double distance = dist(sel,sens);


    string str = (gv_seeds[0])->str();
    for (int i = 1 ; i < gv_nbseeds ; i++)
      str = str + "," + ((gv_seeds[i])->str());
    
    seedproperties e = seedproperties(sel,sens,distance,str);
    l.push_back(e);
    
  } else {


    /*
     * B) compute for all sets of seeds ...
     */

    for (int i=0 ; i < gv_nbseeds ; i++)
      gv_seeds[i] = new seed(gv_minspan,gv_maxspan);
    
    while (1) { /* for all seeds */
          
      automaton * a_spr = NULL;
      double      selp = 0.0;
      
      for(int i=0 ; i < gv_nbseeds ; i++){
	
	/* seed automaton */
	
	automaton * a_s = new automaton();
	SEEDAUTOMATON(*a_s,*gv_seeds[i]);
	
	if (gv_minimize) {
	  automaton * na = a_s->Hopcroft(); 
	  delete a_s; 
	  a_s = na ;
	}
	
	
	/* (1) compute partial selectivity */
	automaton * a_pr_s_sel      = a_s->product(a_sel, false, true);
	selp                       += a_pr_s_sel->PrState(0,gv_seeds[i]->span());	
	delete a_pr_s_sel;
	
	
	/* (2) product automaton for sensitivity */
	
	if (i > 0) {
	  automaton * a_spr_temp = a_spr;
	  a_spr =  a_spr->product(*a_s, false, false);
	  if (gv_minimize) {
	    automaton * na = a_spr->Hopcroft(); 
	    delete a_spr; 
	    a_spr = na ;
	  }
	  delete a_spr_temp;
	  delete a_s;
	} else {
	  a_spr = a_s;
	}
	
      }// for each seed
      
      
      /* (3) compute sensitivity */
      automaton * a_pr_sens     = a_spr->product(a_sens, false, true);
      double sens               = a_pr_sens->PrState(0,gv_alignment_length);	
      delete      a_pr_sens;
      
      double sel = 1.0 - selp;
      
      delete a_spr;
      
      double distance = dist(sel,sens);
      
      
      string str = (gv_seeds[0])->str();
      for (int i = 1 ; i < gv_nbseeds ; i++)
	str = str + "," + ((gv_seeds[i])->str());
      
      seedproperties e = seedproperties(sel,sens,distance,str);
      l.push_back(e);
      
  
      //cout << (e) << endl;

      /* Prune list to keep bets ones */
      if (!(nbruns++ % 10000)) {
	double area = selectPareto(l);
	cerr << "l.size:" << l.size()<< ", area:"<< area << endl; 
      }



      /* B.2 : enumerate the next seed(s) */
      if (gv_nbruns > 0){
	
	/* B.2a Random enumeration */
	for (int i = 0 ; i < gv_nbseeds ; i++ ) {
	  (gv_seeds[i])->random();
	}
	
	if (nbruns >= gv_nbruns)
	  goto label_EndOfLoop;
	
      } else {
	
	/* B.2b Complete enumeration */
	int j = 0;
	while ( (gv_seeds[j])->next() == 0 ) {
	  j++;
	  if (j >= gv_nbseeds)
	    goto label_EndOfLoop;
	}
	
	for(int i = 0 ; i < j ; i++) { // reset the span of previous j-th seeds to min
	  gv_seeds[i] = new seed(gv_minspan,gv_maxspan);
	}

      }/* Random/Complete enum */
    }/* while (1) */
  label_EndOfLoop:
    ;
  } /* enumerate all seeds /or/ check one seed */



  /*
   * III) Display pareto set and area
   */

  cerr << "# seeds\t sel\t sens\t dist"<<  endl;
  list_and_areaPareto(l);

  return 0;
}



